import readline
import __builtin__
import __main__

def unique(inlist):
    outlist = []
    inlist.sort()
    first = 1
    for item in inlist:
        if not first:
            if lastitem == item:
                continue
        lastitem = item
        outlist.append(item)
        first = 0
    return outlist
        
class Completer:

    def __init__(self, global_dict = None, local_dict = None):
        if global_dict is None:
            global_dict = __main__.__dict__
        self.global_dict = global_dict
        if local_dict is None:
            local_dict = global_dict
        self.local_dict = local_dict
        
    def complete(self, text, state):
        if state == 0:
            if "." in text:
                matches = self.attr_matches(text)
            else:
                matches = self.global_matches(text)
            self.matches = unique(matches)
        return self.matches[state]
                
    def global_matches(self, text):
        import keyword
        matches = []
        n = len(text)
        for list in [keyword.kwlist,
                     self.local_dict.keys(),
                     self.global_dict.keys(),
                     __builtin__.__dict__.keys()]:
            for word in list:
                if word[:n] == text:
                    matches.append(word)
        return matches

    def attr_matches(self, text):
        import re
        m = re.match(r"(\w+(\.\w+)*)\.(\w*)", text)
        if not m:
            return []
        expr, attr = m.group(1, 3)
        try:
            object = eval(expr,self.global_dict, self.local_dict)
        except:
            return []
        try:    
            words = dir(object)
        except NameError:
            words = []

        if hasattr(object,'__class__'):
            words.append('__class__')
            words = words + get_class_members(object.__class__)
    
        matches = []
        n = len(attr)
        for word in words:
            if word[:n] == attr:
                matches.append("%s.%s" % (expr, word))
        return matches            


def get_class_members(klass):
    ret = dir(klass)
    if hasattr(klass,'__bases__'):
        for base in klass.__bases__:
            ret = ret + get_class_members(base)
    return ret
