# 
# pyshell
#

import os, sys, math, string, traceback, types, keyword, __builtin__
import Tkinter
from undostack import UndoStack, Undo
from ScrolledText import ScrolledText
from tkFileDialog import asksaveasfilename, asksaveasfilename, \
     askopenfilename
from cPickle import loads, dumps
from string import replace, lower, split, atoi, lstrip, rstrip, \
     strip

from stream import Stream
from interpreter import Interpreter
from completer import Completer

revisionstr = '$Revision: 1.7 $'
revision = revisionstr[11:-1]

sys.ps1 = ">>> "
sys.ps2 = "... "
PROMPT1 = 'prompt1'
PROMPT2 = 'prompt2'
PROTECTED = 'protected'
TEMP    = 'temp'
INPUT   = 'INPUT'
OUTPUT  = 'output'
OUTERR  = 'outerr'
EXCEPTION = 'exception'

OUTINDEX = 'outindex'
INSERT  = 'insert'
END     = 'end'
BEGIN   = '1.0'

open_file = open

class Shell(ScrolledText, UndoStack):

    def __init__(self, parent, namespace, bg='white', wrap='word', 
                 **kw):
        kw['bg'] = bg
        kw['wrap'] = wrap
        self.parent = parent
        
        apply(ScrolledText.__init__,(self, parent), kw)
        apply(UndoStack.__init__, (self,))
        
        self._make_menu()
        self._patch_tk()
        self.counter = 0
        self.clipboard = ''
        self.history_stack = ['']
        self.history_i = -1 
        
        self.bind( '<KeyPress>', self.clearMsg)
        self.bind( '<Return>', self._return)
        self.bind( '<Shift-Return>', self.run)
        self.bind( '<Control-a>', self.goto_eol)
        self.bind( '<Home>', self.goto_eol)
        self.bind( '<Control-q>', self.sendeof)
        self.bind( '<Tab>', self.tab_complete)
        self.bind( '<Control-i>', self.indent) 
        self.bind( '<Control-u>', self.dedent) 
        self.bind( '<Triple-Button>', self.select_line)
        self.bind('<Control-c>', self.copyregion)
        self.bind('<Control-v>', self.paste)
        self.bind('<Control-x>', self.delregion)
        self.bind('<Control-Up>', self.history_prev)
        self.bind('<Control-Down>', self.history_next)
        self.bind('<Control-z>', self._undo)
        self.bind('<Control-r>', self._redo)

        self.tag_config(EXCEPTION, foreground = "red")
        self.tag_config(OUTERR, foreground = "red")
        self.tag_config(OUTPUT, foreground = "black", background = "lightgrey")

        self.tag_config(TEMP, foreground = "blue")
        self.tag_config(PROMPT1, foreground = "blue")
        self.tag_config(PROMPT2, foreground = "blue")

        self.tag_config("sel", background = "yellow")
        self.stderr = Stream(self.write_err, OUTERR, PROTECTED)
        self.stdout = Stream(self.write_out, OUTPUT, PROTECTED) 
        
        
        self.mark_set(OUTINDEX, '1.0')
        self.hpos = 0
        self.tabCount = 0

        self.shell = Interpreter( '<Console>', namespace)
        self.completer = Completer(namespace)

        self.write( 'Python ' +sys.version+'\n', PROTECTED) 
        self.write( 'PyShell ' +revision +'\n\n', PROTECTED)
         
        self.focus_set()
        self.insert_prompt()
        self.clear_undo()

    def _patch_tk(self):
        insert_tcl = self.register(self.insert)
        delete_tcl = self.register(self.delete)

        tclstr = """
        rename _w _old_w

        proc _w { command args } {
            if { $command == "insert" } {
                set ret [ eval "insert_tcl $args" ]
            } elseif {$command == "delete"} {
                set ret [ eval "delete_tcl $args" ]
            } else {
                set ret [ eval "_old_w $command $args" ]
            }
            return $ret
        }
        """
        self._old_w = '.old'+self._w
        tclstr = replace(tclstr, '_old_w', self._old_w)
        tclstr = replace(tclstr, '_w', self._w)
        tclstr = replace(tclstr, 'insert_tcl', insert_tcl)
        tclstr = replace(tclstr, 'delete_tcl', delete_tcl)
        self.tk.eval(tclstr)
        
    def insert(self, index, chars, *tags):
        if self.is_editable(index):
            self.delselection()
            apply(self.forced_insert, (index, chars)+tuple(tags))
            
        else:
            # XXX this is a bit ugly
            # Guess the user want's a new prompt . This can lead to conflicts as the
            # function is also called internally. Use for internal calls _insert or force_
            # insert instead.
            
            _prompt2 = self.tag_nextrange(PROMPT1, 'insert') or (END,END)

            self.mark_set('insert', _prompt2[0]+'-1c')
            self.forced_insert('insert','\n')

            tags = self.tag_names()
            pos = self.index('insert')
            for tag in tags:
                self.tag_remove(tag, pos, 'insert +1c')

            self.insert_prompt()

    def _insert(self, index, chars, *tags):
        if self.is_editable(index):
            #self.delselection()
            apply(self.forced_insert, (index, chars)+tuple(tags))

    def forced_insert(self, index, chars, *tags):
        undo = apply(self._raw_insert, (index, chars)+tags)
        self.add_undo(undo)

    def _raw_insert(self, index, chars, *tags):
        index = self.index(index)
        undo = Undo(self._raw_delete, index, index+'+%i c' % len(chars))

        self.tk.call((self._old_w, 'insert', index, chars) + tuple(tags))
        return undo

    def _raw_delete(self, index1, index2=None):
        index1 = self.index(index1)
        text = self.get(index1, index2)
        undo = Undo(self._raw_insert, index1, text)
        
        self.tk.call(self._old_w, 'delete', index1, index2)
        return undo
        
    def delete(self, index1, index2=None):
        if index2 is None:
            index2 = index1+"+1c"
        if self.is_editable(index1, index2):
            self.forced_delete(index1, index2)

    def forced_delete(self, index1, index2=None):
        undo = self._raw_delete(index1, index2)
        self.add_undo(undo)

    def copyregion(self, event=None):
        range = self.tag_ranges('sel')
        if len(range) !=2:
            return
        self.clipboard = apply(self.get, range)

    def paste(self, event=None):
        self.delselection()
        index = self.index('insert')
        self._insert(index, self.clipboard)
        self.see('insert')
        
    def delregion(self, event=None):
        self.copyregion()
        self.delselection()

    def delselection(self):
        sel = self.tag_ranges('sel')
        if len(sel) == 2:
            apply(self.delete, sel)
            self.tag_remove('sel', sel[0], sel[1])
        
    def _undo(self, event=None):
        self.undo()
        return "break"
        
    def _redo(self, event=None):
        self.redo()
        return "break"
        
    def history_prev(self, event=None):
        i = self.history_i
        if i >= 0:
            
            self.delselection()
            begin = self.index('insert')
            self._insert('insert', self.history_stack[i])
            self.tag_add('sel', begin, 'insert')
    
            if i > 0:
                self.history_i = i-1
        return "break"
        
    def history_next(self, event=None):
        i = self.history_i
        if i < len(self.history_stack): 
    
            self.delselection()
            begin = self.index('insert')
            self._insert('insert', self.history_stack[i])
            self.tag_add('sel', begin, 'insert')
            if i < len(self.history_stack)-1:
                self.history_i = i+1
        return "break"
        
    def _make_menu(self):
        menu = Tkinter.Menu(self)
        
        menu.add_command(label='Load...', command=self.load)
        menu.add_command(label='Save...', command=self.saveas)
        menu.add_command(label='Export...', command=self.export)
        menu.add_command(label='Close', command=self.sendeof)

        self.menu = menu
        self.bind('<Button-3>',
                  lambda e, s=self: s.menu.tk_popup(e.x_root, e.y_root))


    def swapStdFiles(self): 
        sys.stdout,self.stdout = self.stdout,sys.stdout
        sys.stderr,self.stderr = self.stderr,sys.stderr

    def write_out(self, data, *tags):
        apply(self.forced_insert, ('insert', data)+tags)
    write = write_out

    def write_err(self, data, *tags):
        apply(self.forced_insert, ('insert', data)+tags)

    def remove_temp(self):
        ranges = self.tag_ranges(TEMP)
        i = 1
        while i<len(ranges):
            self.delete(ranges[i-1], ranges[i])
            i = i+2

    def Execute(self, input):
        self.swapStdFiles()
        try:
            self.shell.push(input)
        finally:
            self.swapStdFiles()

    def run(self, *args):
        self.remove_temp()
                
        _prompt1 = self.tag_prevrange(PROMPT1, INSERT+' lineend') 
        if len(_prompt1) == 0:
            _prompt1 = self.tag_nextrange(PROMPT1, '1.0')
            # if the text doesn't contain any PROMPT1, take the whole text 
            # as input. This may happen when a textfile is loaded.
            # Otherwise, do nothing.
            if len(_prompt1) != 0:
                return "break"
            _prompt1 = ('1.0', '1.0')
        _prompt2 = self.tag_nextrange(PROMPT2, _prompt1[0]) or (END,END)
        _nextprompt = self.tag_nextrange(PROMPT1, _prompt1[1]) or (END, END)
        
        if self.compare(_prompt2[0], '>', _nextprompt[0]):
            inputend = _nextprompt[0]
        else:
            inputend = _prompt2[0]
            
        #
        # Now get the input text
        input = self.get(_prompt1[1]+"+1c", inputend)
        self.history_stack.append(input)
        self.history_i = len(self.history_stack)-1
        #
        # remove the complete cell
        self.mark_set('insert', _prompt1[0])
        apply(self.forced_delete, ('insert', _nextprompt[0]+"-1c"))

        #
        # insert the "in"-prompt with the new number
        self.counter = self.counter+1
        pos = self.index('insert')
        self.forced_insert('insert', '[In %s]:\n' % self.counter)
        self.tag_add(PROMPT1, pos, pos+" lineend")
        self.tag_add(PROTECTED, pos, pos+" lineend")

        #
        # reinsert the input text
        pos = self.index('insert')
        self.forced_insert('insert', input)
        
        # 
        # depending on the lines before, the text may be taged wrongly
        #  -> remove all tags.
        tags = self.tag_names()
        pos = self.index(pos+'-1l lineend')
        for tag in tags:
            self.tag_remove(tag, pos, 'insert')
        #
        # tag it as INPUT
        self.tag_add(INPUT, pos, 'insert')

        #
        # insert the "out"-prompt        
        pos = self.index('insert')
        self.forced_insert('insert', '[Out %s]:\n' % self.counter)
        self.tag_add(PROMPT2, pos, pos+" lineend")
        self.tag_add(PROTECTED, pos, pos+" lineend")

        self.Execute(input)

        _nextprompt = self.tag_nextrange(PROMPT1, 'insert')
        if len(_nextprompt) == 0:
            self.insert_prompt()
        else:
            self.forced_delete('insert', _nextprompt[0])
            _nextprompt = self.tag_nextrange(PROMPT1, 'insert')
            self.mark_set('insert', _nextprompt[1]+"+1c")
        self.see('insert')

        #
        # I'd love to keep the undo, but this makes things really easier
        self.clear_undo()

        return "break"

    def insert_prompt(self):
        pos = self.index('insert linestart')
        self.forced_insert('insert', '>\n')
        self.tag_add(PROMPT1, pos, pos+" lineend")
        self.tag_add(PROTECTED, pos, pos+" lineend")

    def sendeof(self, *args):
        self.parent.destroy()
        return "break"
        
    def load(self, event=None, name=None):
        if not name:
            name = askopenfilename( \
                filetypes=[("PyShell", "*.pys"), \
                            ("all files", "*")])
        if name:
            self.forced_delete('1.0', 'end')
            self._file_insert(name)
            self.clear_undo()

    def saveas(self, event=None):
        name = asksaveasfilename( \
            filetypes=[("PyShell", "*.pys"), 
                       ("All files", "*")]
            )
        if name:
            self._file_save(name, ('1.0', 'end'), 'pys')

    def export(self, event=None):
        name = asksaveasfilename( \
            filetypes=[("PyShell", "*.pys"), 
                       ("HTML", "*.html"), 
                       ("Python", "*.py"), 
                       ("All files", "*")]
            )
        if name:
            self._file_save(name, ('1.0', 'end'))
        
        
    def _file_insert(self, filename):
        file = open_file(filename, 'r')
        s = file.read()
        file.close()

        if lower(filename[-4:]) == '.pys':
            text, tagging = loads(s)
        else:
            text = s
            tagging = []
            
        self._insert('insert linestart', text)
        # XXX missing: shift indices
        
        for tag, ranges in tagging:
            if len(ranges)>0:
                apply(self.tag_add, (tag,)+ ranges)
                
    def _file_save(self, filename, _range, filetype=None):
        file = open_file(filename, 'w')
        text = apply(self.get, _range)
        
        if filetype is None: # guess from filename 
            if lower(filename[-4:]) == '.pys':
                filetype = 'pys'
            elif lower(filename[-4:]) in ('html', '.htm'):
                filetype = 'html'
            elif lower(filename[-4:]) in ('html', '.htm'):
                filetype = 'py'
            else:
                filetype = 'txt'

        if filetype == 'pys':
            #
            # PyShell format
            
            tagging = []
            for name in self.tag_names():
                if name not in ('sel', ):
                    ranges = self.tag_ranges(name)
                    if len(ranges)>0:
                        tagging.append((name, ranges))
            
            file.write(dumps((text, tagging)))
            file.close()
            
        elif filetype == 'html':
            #
            # html text
            
            file.write('<pre>\n')
            changes = []
            for name in self.tag_names():
                if name in (PROMPT1, PROMPT2, EXCEPTION, TEMP):

                    ranges = self.tag_ranges(name)
                    for i in range(len(ranges)):
                        row, col = map(string.atoi, 
                                       string.split(ranges[i], '.'))
                        changes.append((row, col, i % 2, name))
                        
            changes.sort()

            class pos:
                pass
            pos.row = 1
            pos.col = 0
            
            chunk = ''
            i = -1

            for row, col, odd, name in changes:
                while (pos.row<row) or (pos.col<col and pos.row==row):
                    
                    i = i+1 # move to next char
                    if i >= len(text):    
                        break

                    char = text[i]
                    if char == '\n':
                        pos.row = pos.row+1
                        pos.col = 0
                    if char == '<':
                        file.write('&lt;')
                    else:
                        file.write(char)
                        
                if not odd:
                    file.write('<font color=%s>' \
                        % self.tag_cget(name, "foreground"))
                    
                else:
                    file.write('</font>')
                if i>= len(text):
                    break
                    
            if i<len(text):
                file.write(text[i:])
            file.write('</pre>\n')

        elif filetype == 'py':
            #
            # save as python program
            # XXX missing
            file.write(text)

        else:
            #
            # save as plain text
            file.write(text)
        file.close()
        
                
    def is_editable(self, index, index2=None):
        "return true if text at index is not protected, false otherwise"
        
        if index2 == None:
            range = self.tag_prevrange(PROTECTED, index+"+1c")
            if len(range) != 2:
                return 1
            begin, end = range
            if self.compare(index, '>=', begin) and \
               self.compare(index, '<=', end): 
                return 0
            if self.compare(index, '==', begin+'-1l lineend'):
                return 0
            
            return 1
        else:
            if self.is_editable(index) and self.is_editable(index2):
                next = self.tag_nextrange(PROTECTED, index) or ('end', 'end')
                if self.compare(next[0], '>', index2):
                    return 1
            return 0
                    
    def goto_eol( self, event=None):
        self.mark_set( 'insert', 'insert linestart')
        return "break"

    def select_line( self, event=None):
        self.text.tag_delete( 'sel')
        self.text.tag_add( 'sel', 'insert linestart', 'insert lineend')        
        return "break"

    def clearMsg( self, event=None):
        self.remove_temp()
        self.tabCount = 0
    
        begin = self.tag_prevrange(PROMPT1, INSERT) 
        if len(begin)>0:
            end = self.tag_nextrange(PROMPT2, begin[0]) or (END, END)
            self.tag_add(INPUT, begin[1], end[0])

    def _return( self, event=None):
        self.insert('insert', '\n')
        pos = 'insert-1l'
        lastline = rstrip(self.get(pos, pos+' lineend'))
        n = len(lastline)- len(lstrip(lastline))
        if len(lastline)>0 and lastline[-1] == ':':
            n = n+4
        self.insert('insert', ' '*n)
        self.see('insert')
        return 'break'
    
    def _indent_line(self, pos):
        self._insert(pos+' linestart', ' ' * 4)
    
    def indent( self, event=None):
        sel = self.tag_ranges('sel')
        
        if len(sel) == 2:
            begin = self.index(sel[0]+' linestart')
            end = self.index(sel[1]+' linestart')
            l_begin = atoi(split(begin, '.')[0])
            l_end = atoi(split(end, '.')[0])
            for l in range(l_begin, l_end+1):
                self._indent_line("%i.0" % l)
        else:
            self._indent_line('insert')
        return 'break'

    def _dedent_line(self, pos):
        pos = self.index(pos)
        linestart = self.index(pos+' linestart')
        text = self.get(linestart, pos+' lineend')
            
        maxn = len(text)-len(string.lstrip(text))
        n = min(4, maxn, len(text))
        self.delete(linestart, linestart+'+%ic' % n)

    def dedent( self, event=None):
        sel = self.tag_ranges('sel')
        
        if len(sel) == 2:
            begin = self.index(sel[0]+' linestart')
            end = self.index(sel[1]+' linestart')
            l_begin = atoi(split(begin, '.')[0])
            l_end = atoi(split(end, '.')[0])
            for l in range(l_begin, l_end+1):
                self._dedent_line("%i.0" % l)
        else:
            self._dedent_line('insert')
        return 'break'

    def tab_complete( self, event=None):
        self.tabCount = self.tabCount + 1
        
        index, word = self._getWord()
        words = []
        i = 0
        while 1:
            try:
                alt = self.completer.complete(word, i)
            except IndexError:
                break
            if alt is None: break
            words.append(alt)
            i = i+1
        
        if len( words) == 1:
            self.delete( index, "insert")
            self._insert( "insert", words[0])
            self.tabCount = 0

        elif len( words) > 1:
            self.delete( index, "insert")
            self._insert( "insert", self._getCommon( words))
            
            if self.tabCount == 1:
                self._printTemp( "[Type more to complete or tab to see choices]")
            else:
                self._printTemp( str(words))
            
        elif not words:
            self._printTemp( "[No completion]")
            self.tabCount = 0

        return "break"

    def _getWord(self, index="insert", tail=""):
        
        pos = self.index(index+"-1c")
        char = self.get(pos)
        if char in string.letters+string.digits+"_.":
            return self._getWord(pos, char+tail)
        else:
            return index, tail


    def _printTemp( self, msg):
        self.remove_temp()
        index = self.index( "insert")
        tmp = 'tmp'
        self.mark_set(tmp, index+" lineend")
        self._insert( tmp, "\n" + msg, TEMP) 
        self.mark_set( "insert", index)
        self.see(tmp)
            
    def _getCommon( self, words):
        base = ""
        pos = 0
        
        try:
            while 1:
                c = words[0][pos]
                for word in words:
                    if word[pos] != c: return base
                base = base + c
                pos = pos + 1
        except IndexError:
            pass
            
        return base
                        
class ShellWindow(Tkinter.Toplevel):
    
    def __init__(self, master, namespace, cnf={}, **kw):
        apply(Tkinter.Toplevel.__init__, (self,master,cnf),kw)
        shell = Shell(self, namespace)
        shell.pack(fill = 'both', expand = 1)
        self.bind( '<Escape>', self.destroy)

def open(namespace = None):
    app = Tkinter.Tk()
    if namespace is None:
        namespace = {}
    else:
        namespace = namespace.copy()
    namespace['Tkinter'] = Tkinter
    namespace['string'] = string
    namespace['math'] = math
    namespace['sys'] = sys
    namespace['os'] = os

    shell = Shell(app, namespace, width = 40, font=('courier', 12))
    shell.pack(fill = 'both', expand = 1)
    return shell

if __name__ == "__main__":
    shell = open()
    if len(sys.argv) > 1:
        shell.load(name=sys.argv[1])
    
    shell.mainloop()
    
